#include "stdafx.h"
#include "template.h"

#include "minmax.h"
#include "myfile.h"


#pragma pack(1)
struct dialog_template_struct
	{
	long dtStyle;
	BYTE dtItemCount;
	WORD dtX, dtY;
	WORD dtCX, dtCY;
	};
#pragma pack()

struct dialog_buf_item_struct
	{
	int dtilX, dtilY;
	int dtilCX, dtilCY;
	unsigned int dtilID;
	long dtilStyle;
	};
	
#ifdef _DEBUG
	// Define this to get debugging trace output
	// #define DEBUG
	#undef THIS_FILE
	static char BASED_CODE THIS_FILE[] = __FILE__;
	#define new DEBUG_NEW
#endif

	
CDialogItem::CDialogItem(CDialogTemplate *pTemplate, LPCSTR szClass, long lStyle)
{
	m_rect.SetRect(-1, -1, -1, -1);
	m_uID = 0;
	m_lStyle = lStyle | WS_CHILD | WS_VISIBLE;
	m_szClass = szClass;
	m_szText.Empty();
	m_pTemplate = pTemplate;
}


CDialogItem::CDialogItem(CDialogTemplate *pTemplate, unsigned char cClass, long lStyle)
{
	m_rect.SetRect(-1, -1, -1, -1);
	m_uID = 0;
	m_szClass = (char)cClass;
	m_szText.Empty();
	m_pTemplate = pTemplate;

	if (lStyle != LONG_MIN)
		m_lStyle = lStyle | WS_CHILD | WS_VISIBLE;
	else switch (cClass)
		{
		case CLASS_BUTTON :
			m_lStyle = BS_PUSHBUTTON | WS_TABSTOP | WS_CHILD | WS_VISIBLE;
			break;
		case CLASS_EDIT :
			m_lStyle = ES_LEFT | WS_TABSTOP | WS_CHILD | WS_VISIBLE;
			break;
		case CLASS_STATIC :
			m_lStyle = SS_LEFT | SS_NOPREFIX | WS_CHILD | WS_VISIBLE;
			break;
		case CLASS_LISTBOX :
			m_lStyle = WS_TABSTOP | WS_CHILD | WS_VISIBLE;
			break;
		case CLASS_COMBOBOX :
			m_lStyle = WS_TABSTOP | CBS_DROPDOWN | WS_CHILD | WS_VISIBLE;
			break;
		case CLASS_SCROLLBAR :
		default :
			m_lStyle = WS_CHILD | WS_VISIBLE;
			break;
		}
}


CDialogItem::CDialogItem(CDialogTemplate *pTemplate, const char *pData, int *piBytesUsed)
{
	int iUsed, iLen, nData;
	const struct dialog_buf_item_struct *pItem;

#ifdef DEBUG
	TRACE("CDialogItem::CDialogItem(pTemplate, pData, piBytesUsed) entered [this=%04X:%04X]\n",
		SELECTOROF(this), OFFSETOF(this));
#endif

	m_pTemplate = pTemplate;
	
	pItem = (const struct dialog_buf_item_struct *)pData;
	pData += sizeof(struct dialog_buf_item_struct);
	iUsed = sizeof(struct dialog_buf_item_struct);

	m_uID = pItem->dtilID;
	m_lStyle = pItem->dtilStyle;
#ifdef DEBUG
	TRACE("   iId = %04X; lStyle = %08lX\n", m_uID, m_lStyle);
#endif

	m_rect.left = pItem->dtilX;
	m_rect.top = pItem->dtilY;
	m_rect.right = pItem->dtilX + pItem->dtilCX;
	m_rect.bottom = pItem->dtilY + pItem->dtilCY;

	// pData points to the class right now!
	if (pData[0] & 0x80)
		{
		m_szClass = pData[0];
		pData++;
		iUsed++;
		}
	else
		{
		m_szClass = pData;
		iLen = lstrlen(pData);
		pData += iLen + 1;
		iUsed += iLen + 1;
		}

	m_szText = pData;
	iLen = lstrlen(pData) + 1;
	pData += iLen;
	iUsed += iLen;
#ifdef DEBUG
	TRACE("   szText = \"%s\"\n", (LPCSTR)m_szText);
#endif

    // Check for extra data but toss it for now ...
	nData = pData[0];
	pData++;
	iUsed++;

	if (nData)
		iUsed += nData;

	*piBytesUsed = iUsed;
}

		
CDialogItem::CDialogItem(CDialogTemplate *pTemplate, const CDialogItem *pSrcItem)
{
#ifdef DEBUG
	TRACE("CDialogItem::CDialogItem(pTemplate=%04X:%04X, pSrcItem=%04X:%04X) entered [this=%04X:%04X]\n",
		SELECTOROF(pTemplate), OFFSETOF(pTemplate),
		SELECTOROF(pSrcItem), OFFSETOF(pSrcItem),
		SELECTOROF(this), OFFSETOF(this));
#endif

	ASSERT(pTemplate != NULL);
	ASSERT(pSrcItem != NULL);

	m_rect = pSrcItem->m_rect;
	m_uID = pSrcItem->m_uID;
	m_lStyle = pSrcItem->m_lStyle;
	m_iFocusPriority = pSrcItem->m_iFocusPriority;
	m_szClass = pSrcItem->m_szClass;
	m_szText = pSrcItem->m_szText;
	m_pTemplate = pTemplate;
}


CPoint CDialogItem::get_position(void)
{
	CPoint pt;

	pt.x = m_rect.left;
	pt.y = m_rect.top;

	return pt;
}


CDialogItem *CDialogItem::set_size(int iWidth, int iHeight)
{
	m_rect.right = m_rect.left + iWidth;
	m_rect.bottom = m_rect.top + iHeight;

	return this;
}


CDialogItem *CDialogItem::set_text(LPCSTR szText)
{
	CSize size;
	
	m_szText = szText;
	
	if ( (m_rect.bottom == -1) && (m_rect.right == -1) && (m_szClass[0] & 0x80) )
		{
		switch (m_szClass[0])
			{
			case CLASS_BUTTON :
				{
				int iType = (int)(m_lStyle & 0x0F);
				
				size = m_pTemplate->pixel_to_dialog(m_pTemplate->get_dc()->GetTextExtent(szText, lstrlen(szText)));
				if ( (iType == BS_PUSHBUTTON) || (iType == BS_DEFPUSHBUTTON) )
					set_size(max(size.cx + 6, 36), max(size.cy + 6, 14));
				else if ( (iType == BS_RADIOBUTTON) || (iType == BS_AUTORADIOBUTTON) )
					set_size(size.cx + 10, max(size.cy + 2, 10));
				else if ( (iType == BS_CHECKBOX) || (iType == BS_AUTOCHECKBOX) )
					set_size(size.cx + 10, max(size.cy + 2, 10));
				}
				break;
			case CLASS_EDIT :
				size = m_pTemplate->pixel_to_dialog(m_pTemplate->get_dc()->GetTextExtent(szText, lstrlen(szText)));
				set_size(size.cx + 4, size.cy + 4);
				break;
			case CLASS_STATIC :
				size = m_pTemplate->pixel_to_dialog(m_pTemplate->get_dc()->GetTextExtent(szText, lstrlen(szText)));
				set_size(size.cx, size.cy);
				break;
			case CLASS_LISTBOX :
				size = m_pTemplate->pixel_to_dialog(m_pTemplate->get_dc()->GetTextExtent(szText, lstrlen(szText)));
				set_size(size.cx, size.cy);
				break;
			case CLASS_COMBOBOX :
				// Height = 12 + 8 * nVisibleItems
				size = m_pTemplate->pixel_to_dialog(m_pTemplate->get_dc()->GetTextExtent(szText, lstrlen(szText)));
				set_size(size.cx, size.cy);
				break;
			case CLASS_SCROLLBAR :
			default :
				break;
			}
		}
		
	return this;
}


BOOL CDialogItem::generate(CFile &hFile)
{
	struct dialog_buf_item_struct sHeader;
	int iLen;

    if ( (m_rect.IsRectEmpty())
    	|| (m_szClass.IsEmpty())
    	|| (m_rect.left < 0)
    	|| (m_rect.top < 0) )
    	return FALSE;

	sHeader.dtilX = m_rect.left;
	sHeader.dtilY = m_rect.top;
	sHeader.dtilCX = m_rect.Width();
	sHeader.dtilCY = m_rect.Height();
	sHeader.dtilID = m_uID;
	sHeader.dtilStyle = m_lStyle;

	hFile.Write(&sHeader, sizeof(sHeader));
	
	if (m_szClass[0] & 0x80)
		iLen = 1;
	else
		iLen = m_szClass.GetLength() + 1;
	hFile.Write((LPCSTR)m_szClass, iLen);

	iLen = 0;
	
	if (m_szText.IsEmpty())
		hFile.Write(&iLen, 1);
	else
		hFile.Write((LPCSTR)m_szText, m_szText.GetLength() + 1);

	// No extra data
	hFile.Write(&iLen, 1);
	
	return TRUE;
}
	

CDialogTemplate::CDialogTemplate(CWnd *pParent)
{
	CWinApp *pApp;

#ifdef DEBUG
	TRACE("CDialogTemplate::CDialogTemplate(%04X:%04X) entered [this=%04X:%04X]\n",
		SELECTOROF(pParent), OFFSETOF(pParent), SELECTOROF(this), OFFSETOF(this));
#endif

	m_lStyle = WS_BORDER | WS_DLGFRAME | WS_POPUP | WS_VISIBLE | WS_CLIPSIBLINGS
		| DS_MODALFRAME | WS_SYSMENU;
	m_rect.SetRectEmpty();
	m_lDialogUnits = GetDialogBaseUnits();
	m_pParent = pParent;

	m_pDC = NULL;
	if (m_pParent)
		{
		m_pDC = new CClientDC(m_pParent);
		ASSERT_VALID(m_pDC);
		}
	else
		{
		pApp = AfxGetApp();
		if (pApp)
			{
			if (pApp->m_pMainWnd)
				{
				m_pDC = new CClientDC(pApp->m_pMainWnd);
				ASSERT_VALID(m_pDC);
				}
			}
		}

	m_pFont = NULL;
}
	

CDialogTemplate::~CDialogTemplate()
{
	CDialogItem *pItem;

#ifdef DEBUG
	TRACE0("CDialogTemplate::~CDialogTemplate() entered\n");
#endif

	while (m_aItems.GetSize())
		{
		pItem = (CDialogItem *)m_aItems.GetAt(0);
		delete pItem;
		m_aItems.RemoveAt(0);
		}
	if (m_pDC)
		delete m_pDC;
	if (m_pFont)
		delete m_pFont;
}


BOOL CDialogTemplate::load(LPCSTR szTemplateName)
{
	tResource hResource;
	
#ifdef DEBUG
	if (SELECTOROF(szTemplateName))
		TRACE("CDialogTemplate::load(\"%s\") entered\n", szTemplateName);
	else
		TRACE("CDialogTemplate::load(%04X) entered\n", OFFSETOF(szTemplateName));
#endif

	if (!hResource.find(szTemplateName, RT_DIALOG))
		return FALSE;

	return load(hResource);
}


BOOL CDialogTemplate::load(int iTemplateID)
{
	tResource hResource;
	
#ifdef DEBUG
	TRACE("CDialogTemplate::load(%04X) entered\n", iTemplateID);
#endif

	if (!hResource.find(iTemplateID, RT_DIALOG))
		return FALSE;

	return load(hResource);
}


BOOL CDialogTemplate::load(tResource &hResource)
{
	char *pData;
	int i, iUsed;
	struct dialog_template_struct *pHeader;
	WORD wPointSize;
	CString szFaceName;
	CDialogItem *pItem;

#ifdef DEBUG
	TRACE("CDialogTemplate::load(hResource) entered at %08ld\n", GetTickCount());
#endif

	VERIFY(hResource.load());

	pData = (char *)hResource.lock();
	ASSERT(pData);

	// Extract header
	pHeader = (struct dialog_template_struct *)pData;
	pData += sizeof(struct dialog_template_struct);
	m_lStyle = pHeader->dtStyle;
	m_rect.left = pHeader->dtX;
	m_rect.top = pHeader->dtY;
	m_rect.right = m_rect.left + pHeader->dtCX;
	m_rect.bottom = m_rect.top + pHeader->dtCY;

	// Get menu name
	m_szMenuName = pData;
	pData += lstrlen(pData) + 1;
#ifdef DEBUG
	TRACE("   Menu name \"%s\"\n", (LPCSTR)m_szMenuName);
#endif

	// Get class name
	m_szClassName = pData;
	pData += lstrlen(pData) + 1;
#ifdef DEBUG
	TRACE("   Class name \"%s\"\n", (LPCSTR)m_szClassName);
#endif

	// Get caption text
	m_szCaption = pData;
	pData += lstrlen(pData) + 1;
#ifdef DEBUG
	TRACE("   Caption text \"%s\"\n", (LPCSTR)m_szCaption);
#endif
    
    // Get font, if specified
    if (m_lStyle & DS_SETFONT)
    	{
    	wPointSize = *((WORD *)pData);
    	pData += sizeof(WORD);
    	szFaceName = pData;
    	pData += lstrlen(pData) + 1;
    	set_font(szFaceName, wPointSize);
    	}
    else
    	{
    	m_wPointSize = 0;
    	m_szFaceName.Empty();
    	}
	
	// Load items
	for (i = 0 ; i < (int)pHeader->dtItemCount ; i++)
		{
		pItem = new CDialogItem(this, pData, &iUsed);
		m_aItems.Add(pItem);
		pData += iUsed;
		}

	// Done with the resource - unlock it
	hResource.unlock();

#ifdef DEBUG
	TRACE("CDialogTemplate::load() complete at %08ld\n", GetTickCount());
#endif

	return TRUE;
}


CDialogTemplate & CDialogTemplate::set_size(int iWidth, int iHeight)
{
	m_rect.right = m_rect.left + iWidth;
	m_rect.bottom = m_rect.top + iWidth;
	
	return *this;
}


CSize CDialogTemplate::get_size(void)
{
	CSize size;

	size.cx = m_rect.Width();
	size.cy = m_rect.Height();

	return size;
}


CDialogItem *CDialogTemplate::add_item(LPCSTR szClass, long lStyle)
{
	CDialogItem *pItem;
	
	pItem = new CDialogItem(this, szClass, lStyle);
	
	m_aItems.Add(pItem);
	
	return pItem;
}


CDialogItem *CDialogTemplate::add_item(unsigned char cClass, long lStyle)
{
	CDialogItem *pItem;
	
	pItem = new CDialogItem(this, cClass, lStyle);
	
	m_aItems.Add(pItem);
	
	return pItem;
}


CDialogItem *CDialogTemplate::add_item(const CDialogItem *pItemToCopy, int iIndex)
{
	CDialogItem *pItem;

	pItem = new CDialogItem(this, pItemToCopy);

	if (iIndex < 0)
		m_aItems.Add(pItem);
	else
		m_aItems.InsertAt(iIndex, pItem);

	return pItem;
}


BOOL CDialogTemplate::set_font(LPCSTR szFaceName, WORD wPointSize)
{
	LOGFONT sFont;
	TEXTMETRIC sMetric;

	m_wPointSize = wPointSize;
	m_szFaceName = szFaceName;

	if (!m_pDC)
		return FALSE;

	ASSERT_VALID(m_pDC);
	ASSERT(szFaceName != NULL);
	ASSERT(wPointSize > 0);
	
	sFont.lfHeight = -MulDiv(wPointSize, m_pDC->GetDeviceCaps(LOGPIXELSY), 72);
	sFont.lfWidth = 0;
	sFont.lfEscapement = 0;
	sFont.lfOrientation = 0;
	sFont.lfWeight = FW_DONTCARE;
	sFont.lfItalic = FALSE;
	sFont.lfUnderline = FALSE;
	sFont.lfStrikeOut = 0;
	sFont.lfCharSet = ANSI_CHARSET;
	sFont.lfOutPrecision = OUT_DEFAULT_PRECIS;
	sFont.lfClipPrecision = CLIP_DEFAULT_PRECIS;
	sFont.lfQuality = DEFAULT_QUALITY;
	sFont.lfPitchAndFamily = DEFAULT_PITCH;
	lstrcpy(sFont.lfFaceName, szFaceName);

	m_pFont = new CFont;
		
	if (!m_pFont->CreateFontIndirect(&sFont))
		{
		delete m_pFont;
		return FALSE;
		}

	// Stick font into the DC to get the text metrics, and leave it in
	// so we can get text sizes in the individual controls
	m_pDC->SelectObject(m_pFont);
	m_pDC->GetTextMetrics(&sMetric);

	// m_lDialogUnits = MAKELONG(sMetric.tmAveCharWidth, sMetric.tmHeight);

	return TRUE;
}


void *CDialogTemplate::generate(UINT *puTemplateSize)
{
	CMyMemFile hFile;
	int i;
	CDialogItem *pItem;
	struct dialog_template_struct sHeader;
	void *pData;

#ifdef DEBUG
	TRACE("CDialogTemplate::generate(%04X:%04X) entered\n",
		SELECTOROF(puTemplateSize), OFFSETOF(puTemplateSize));
#endif

	if (puTemplateSize)
		*puTemplateSize = 0;

	if (m_rect.IsRectEmpty())
		return NULL;
		
	sHeader.dtStyle = m_lStyle;
	sHeader.dtItemCount = m_aItems.GetSize();
	sHeader.dtX = m_rect.left;
	sHeader.dtY = m_rect.top;
	sHeader.dtCX = m_rect.Width();
	sHeader.dtCY = m_rect.Height();
	
	hFile.Write(&sHeader, sizeof(sHeader));
	
	hFile.Write((LPCSTR)m_szMenuName, m_szMenuName.GetLength() + 1);
	hFile.Write((LPCSTR)m_szClassName, m_szClassName.GetLength() + 1);
	hFile.Write((LPCSTR)m_szCaption, m_szCaption.GetLength() + 1);
	
	if (m_lStyle & DS_SETFONT)
		{
		hFile.Write(&m_wPointSize, sizeof(WORD));
		hFile.Write((LPCSTR)m_szFaceName, m_szFaceName.GetLength() + 1);
		}
		
	for (i = 0 ; i < num_items() ; i++)
		{
		pItem = get_item(i);
		if (!pItem->generate(hFile))
			return NULL;
		}

	if (puTemplateSize)
		*puTemplateSize = (UINT)hFile.GetLength();

	pData = hFile.Detach();

#ifdef DEBUG
	TRACE("   Returning pointer %04X:%04X\n", SELECTOROF(pData), OFFSETOF(pData));
#endif

	return pData;
}
	

int CDialogTemplate::find_item_index(UINT uID)
{
	int i;
	CDialogItem *pItem;

	for (i = 0 ; i < num_items() ; i++)
		{
		pItem = get_item(i);
		if (pItem->get_id() == uID)
			return i;
		}

	return -1;
}


CDialogItem *CDialogTemplate::find_item(UINT uID)
{
	int iIndex = find_item_index(uID);

	if (iIndex >= 0)
		return get_item(iIndex);
	else
		return NULL;
}


BOOL CDialogTemplate::remove_item(UINT uID)
{
	int i;
	CDialogItem *pItem;

	for (i = 0 ; i < num_items() ; i++)
		{
		pItem = get_item(i);
		if (pItem->get_id() == uID)
			{
			m_aItems.RemoveAt(i);
			delete pItem;
			return TRUE;
			}
		}

	return FALSE;
}


UINT CDialogTemplate::unused_id(void)
{
	int i;
	UINT uID;
	CDialogItem *pItem;

	for (uID = 0x1000 ; uID < 0x7FFF ; uID++)
		{
		for (i = 0 ; i < m_aItems.GetSize() ; i++)
			{
			pItem = (CDialogItem *)m_aItems[i];
			ASSERT(pItem != NULL);
			if (pItem->get_id() == uID)
				break;
			}
		if (i >= m_aItems.GetSize())
			return uID;
		}

	// Hopefully we never get here
	ASSERT(0);
	return 0;
}


int CDialogTemplate::pixel_x_to_dialog(int iPixelX)
{
  int iDialogX;

  if (iPixelX < 0)
    iDialogX = -1;
  else
    iDialogX = (iPixelX * 4 + LOWORD(m_lDialogUnits) / 2) / LOWORD(m_lDialogUnits);

#ifdef DEBUG
  TRACE("pixel_x_to_dialog() : %d -> %d\n", iPixelX, iDialogX);
#endif

  return iDialogX;
}


int CDialogTemplate::pixel_y_to_dialog(int iPixelY)
{
  int iDialogY;

  if (iPixelY < 0)
    iDialogY = -1;
  else
    iDialogY = (iPixelY * 8 + HIWORD(m_lDialogUnits) / 2) / HIWORD(m_lDialogUnits);

#ifdef DEBUG
	TRACE("pixel_y_to_dialog() : %d -> %d\n", iPixelY, iDialogY);
#endif

	return iDialogY;
}


CSize CDialogTemplate::pixel_to_dialog(const CSize &size)
{
	CSize newsize;
	
	newsize.cx = pixel_x_to_dialog(size.cx);
	newsize.cy = pixel_y_to_dialog(size.cy);

	return newsize;
}


CRect CDialogTemplate::pixel_to_dialog(const CRect &rect)
{
	CRect newRect;
	
	newRect.left = pixel_x_to_dialog(rect.left);
	newRect.top = pixel_y_to_dialog(rect.top);
	newRect.right = pixel_x_to_dialog(rect.right);
	newRect.bottom = pixel_y_to_dialog(rect.bottom);

	return newRect;
}


int CDialogTemplate::dialog_x_to_pixel(int x)
{
  int iPixelX;

  if (x < 0)
    iPixelX = -1;
  else
    iPixelX = (x * LOWORD(m_lDialogUnits) + 2) / 4;

#ifdef DEBUG
  TRACE("dialog_x_to_pixel() : %d -> %d\n", x, iPixelX);
#endif

  return iPixelX;
}


int CDialogTemplate::dialog_y_to_pixel(int y)
{
	int iPixelY;

	if (y < 0)
		iPixelY = -1;
	else
		iPixelY = (y * HIWORD(m_lDialogUnits) + 4) / 8;

#ifdef DEBUG
	TRACE("dialog_y_to_pixel() : %d -> %d\n", y, iPixelY);
#endif

	return iPixelY;
}


CSize CDialogTemplate::dialog_to_pixel(const CSize &size)
{
	CSize newsize;
	
	newsize.cx = dialog_x_to_pixel(size.cx);
	newsize.cy = dialog_y_to_pixel(size.cy);
	
	return newsize;
}


CRect CDialogTemplate::dialog_to_pixel(const CRect &rect)
{
	CRect newRect;
	
	newRect.left = dialog_x_to_pixel(rect.left);
	newRect.top = dialog_y_to_pixel(rect.top);
	newRect.right = dialog_x_to_pixel(rect.right);
	newRect.bottom = dialog_y_to_pixel(rect.bottom);

	return newRect;
}

